import logging
import time
import numpy as np
import yaml

from modestga import ga
from modestga.benchmark.methods import differential_evolution
from modestga.benchmark.methods import monte_carlo
from modestga.benchmark.functions import rastrigin

logging.basicConfig(level='INFO')


if __name__ == '__main__':
    # Test name
    tname = 'test_3'  # Change name and parameters below

    # Number of tests
    n_tests = 5

    # Number of dimensions
    n_dims = [2, 4, 8, 16, 32, 64, 128]

    # Functions to test
    functions = [
        ga.minimize,
        differential_evolution.minimize,
        monte_carlo.minimize
    ]

    # Options
    options = {
        'generations': 1000,  # Max. number of generations
        'pop_size': 100,      # Population size
        'tol': 1e-3           # Solution tolerance
    }

    mutation = {
        'ga': 0.0025,  # Initial mutation, it's adaptive
        'differential_evolution': (0, 0.5),
        'monte_carlo': None
    }

    # Results
    res = dict()

    for i in range(n_tests):
        k1 = f'run={i}'
        res[k1] = dict()

        for nd in n_dims:
            k2 = f'ndim={nd}'
            res[k1][k2] = dict()

            for fun in functions:
                fname = fun.__module__.split('.')[-1]
                k3 = f'method={fname}'
                res[k1][k2][k3] = dict()

                mut = mutation[fname]
                print(mut, fname)
                opts = options.copy()
                if mut is not None:
                    opts['mut_rate'] = mut

                bounds = [(-5.12, 5.12) for i in range(nd)]
                t0 = time.perf_counter()
                out = fun(rastrigin, bounds, x0=None, options=opts)
                elapsed_t = time.perf_counter() - t0

                res[k1][k2][k3]['f(x)'] = float(out.fx)
                res[k1][k2][k3]['x'] = out.x.tolist()
                res[k1][k2][k3]['nfev'] = int(out.nfev)
                res[k1][k2][k3]['ng'] = int(out.ng)
                res[k1][k2][k3]['message'] = str(out.message)
                res[k1][k2][k3]['time'] = elapsed_t
                res[k1][k2][k3]['mut'] = f"{mut}"


                with open(f'modestga/benchmark/results/{tname}.yaml', 'w') as yaml_file:
                    yaml.dump(res, yaml_file)

    print('Finished!')



"""This parallel mode is DEPRECATED. Use modestga.minimize() instead."""
from multiprocessing import Process, Queue
import queue
import os
import logging
import time

import pandas as pd
import numpy as np
import cloudpickle

import modestga
from modestga.ga import OptRes
from modestga.benchmark.functions import rastrigin

logging.basicConfig(level='INFO', format="[%(processName)s][%(levelname)s] %(message)s")


class SimpleParallel:
    """Runs multiple minimization processes in parallel and returns all results.

    This is the most simple implementation in which there is no communication
    between processes. Instead, the best final solution is returned.
    """
    def __init__(self, fun, bounds, x0=None, args=(),
                 callback=None, options={}, workers=os.cpu_count() - 1):

        self.pickled_fun = cloudpickle.dumps(fun)
        self.bounds = bounds
        self.x0 = x0
        self.pickled_args = cloudpickle.dumps(args)
        self.pickled_callback = cloudpickle.dumps(callback)
        self.options = options
        self.workers = workers
        self.queue = Queue()

    def run(self):
        processes = list()
        all_results = list()
        for i in range(self.workers):
            pid = i
            p = Process(target=SimpleParallel._optimization_process,
                        args=(self.pickled_fun,
                              self.bounds,
                              self.x0,
                              self.pickled_args,
                              self.pickled_callback,
                              self.options,
                              self.queue,
                              pid)
            )
            p.start()
            processes.append(p)

        # Wait for the results
        n_results = 0

        while n_results < self.workers:
            try:
                res = self.queue.get(True, timeout=5)
                all_results.append(res)
                n_results = len(all_results)
            except queue.Empty:
                time.sleep(0.05)
                continue

        # Join processes
        for p in processes:
            p.join()

        # All results to dataframe
        df = pd.DataFrame(all_results).sort_values('fx')

        logging.info(f"Results from all processes:\n{df}")

        return df

    @staticmethod
    def _optimization_process(pickled_fun, bounds, x0,
                              pickled_args, pickled_callback, options,
                              queue, pid):
        fun = cloudpickle.loads(pickled_fun)
        args = cloudpickle.loads(pickled_args)
        callback = cloudpickle.loads(pickled_callback)

        result = modestga.minimize(fun, bounds, x0, args, callback, options, workers=1)

        result_dict = {
            'pid': pid,
            'x': result.x,
            'message': result.message,
            'ng': result.ng,
            'nfev': result.nfev,
            'fx': result.fx
        }

        queue.put(result_dict, block=True, timeout=5)


def minimize(fun, bounds, x0=None, args=(), callback=None, options={},
             workers=os.cpu_count() - 1):
    """Wrapper over SimpleParallel providing same interface as modestga.minimize()"""

    pmin = SimpleParallel(fun, bounds, x0, args, callback, options, workers)
    res = pmin.run()
    best_res = res.iloc[0]  # Beacuse df is sorted by fx in ascending order

    res = OptRes(
        x = best_res['x'],
        message = best_res['message'],
        ng = best_res['ng'],
        nfev = best_res['nfev'],
        fx = best_res['fx']
    )
    return res


if __name__ == "__main__":
    # Example
    from modestga.benchmark.functions import rastrigin
    fun = rastrigin
    bounds = [(-5.12, 5.12) for i in range(64)]
    options = {
        'generations': 100,
        'pop_size': 100,
        'tol': 1e-3
    }
    def callback(x, fx, ng, *args):
        """Callback function called after each generation"""
        # print(f"\nCallback example:\nx=\n{x}\nf(x)={fx}\n")
        pass

    res = minimize(fun, bounds, callback=callback, options=options)
    print(res)



import logging
from scipy.optimize import differential_evolution
import numpy as np
from modestga.ga import OptRes


def minimize(fun, bounds, x0=None, args=(), callback=None, options={}):
    """Differential Evolution from scipy.

    This is a thin wrapper over the scipy implementation::

        scipy.optimize.differential_evolution(
            func, bounds, args=(), strategy='best1bin', maxiter=1000,
            popsize=15, tol=0.01, mutation=(0.5, 1), recombination=0.7,
            seed=None, callback=None, disp=False, polish=True,
            init='latinhypercube', atol=0, updating='immediate',
            workers=1, constraints=()
        )

    Default options::

        options = {
            'generations': 1000,
            'pop_size': 100,
            'tol': 1e-3,
            'polish': False
        }

    :param fun: function to be minimized
    :param bounds: tuple, parameter bounds
    :param x0: numpy 1D array, initial parameters
    :param args: tuple, positional arguments to be passed to `fun`
    :param callback: function, called after every generation
    :param options: dict, DE options
    :return: OptRes, optimization result
    """
    np.set_printoptions(precision=3)
    log = logging.getLogger(name='minimize(DE)')

    opts = {
        'generations': 1000,
        'pop_size': 100,
        'tol': 1e-3,
        'polish': False,
        'mut_rate': (0.5, 1)
    }

    for k in options:
        if k in opts:
            log.info('Override default option: {}={}'.format(k, options[k]))
            opts[k] = options[k]
        else:
            raise KeyError("Option '{}' not found".format(k))

    res_de = differential_evolution(
        fun,
        bounds,
        args,
        maxiter=opts['generations'],
        popsize=opts['pop_size'],
        tol=opts['tol'],
        mutation=opts['mut_rate'],
        polish=False,
        disp=True,
        workers=1
    )

    res = OptRes(
        x=res_de.x,
        message=res_de.message,
        ng=res_de.nit,
        nfev=res_de.nfev,
        fx=res_de.fun
    )

    return res


if __name__ == "__main__":
    # Example
    logging.basicConfig(level='INFO')

    from modestga.benchmark.functions import rastrigin

    N = 50
    bounds = [(-5.12, 5.12) for i in range(N)]
    res = minimize(rastrigin, bounds, x0=None)
    print(res)



import logging
import numpy as np


def bounds_tuples_to_array(t):
    """
    Convert a list of tuples with bounds into an array with the first
    dimension for lower bounds and the second for upper bounds.
    """
    a = np.array([
        [b[0] for b in t],
        [b[1] for b in t]
    ])
    return a


def bounds_array_to_tuples(a):
    """
    Inversed conversion to `bounds_tuples_to_array()`.
    """
    t = tuple([(lo, hi) for lo, hi in zip(a[0], a[1])])
    return t


class Individual():

    # Total number of instances
    count = 0

    def __init__(self, genes, bounds, fun, args=(), val=None):
        """
        :param genes: 1D array, floats between 0 and 1 (inclusive)
        :param bounds: tuple of tuples, parameter bounds (min, max)
        :param fun: function to be minimized
        :param args: arguments to pass to fun
        :param val: Function value or None
        """
        # Set name
        self.id = 'Ind#{}'.format(Individual.count)
        self.log = logging.getLogger(name=self.id)
        Individual.count += 1

        # Copy arguments
        self.gen = np.array(genes)
        self.fun = fun
        self.args = args
        self.bnd = bounds_tuples_to_array(bounds)

        # Evaluate and save score
        if val is None:
            self.val = self.evaluate()
        else:
            self.val = val

    def set_genes(self, g):
        self.gen = g

    def get_estimates(self):
        return self.bnd[0] + self.gen * (self.bnd[1] - self.bnd[0])

    def evaluate(self):
        """
        Evaluate cost function.

        Instead of calling this method,
        you may read the instance attribute `val`.
        """
        self.val = self.fun(self.get_estimates(), *self.args)
        return self.val

    def copy(self):
        ind = Individual(
            self.gen,
            bounds_array_to_tuples(self.bnd),
            self.fun,
            self.args,
            self.val
        )
        return ind

    def __str__(self):
        s = '{}: {} -> {}'.format(self.id, self.gen, self.val)
        return s



import logging
import numpy as np
from modestga.ga import OptRes


def minimize(fun, bounds, x0=None, args=(), callback=None, options={}):
    """Monte Carlo optimization.

    The options have similar names to GA to avoid confusion.
    The default options are::

        options = {
            'generations': 10000, # Max. number of generations (iterations)
            'pop_size': 1000,     # Population size (number of guesses per iteration)
            'tol': 1e-3,          # Solution tolerance
            'inertia': 1000,      # Max. number of non-improving generations
        }

    :param fun: function to be minimized
    :param bounds: tuple, parameter bounds
    :param x0: numpy 1D array, initial parameters
    :param args: tuple, positional arguments to be passed to `fun`
    :param callback: function, called after every generation
    :param options: dict, MC options
    :return: OptRes, optimization result
    """
    np.set_printoptions(precision=3)
    log = logging.getLogger(name='minimize(MC)')

    def norm(x: np.array, b: np.array) -> np.array:
        """Normalizes x with respect to bounds."""
        return b[:, 0] + x * (b[:, 1] - b[:, 0])

    def random_x(b: np.array):
        """Returns random parameters within bounds `b`."""
        rx = norm(
            np.random.uniform(low=0., high=1., size=bounds.shape[0]),
            b)
        return rx

    opts = {
        'generations': 1000,  # Max. number of generations
        'pop_size': 100,      # Population size
        'tol': 1e-3,          # Solution tolerance
        'inertia': 1000,      # Max. number of non-improving generations
    }

    for k in options:
        if k in opts:
            log.info('Override default option: {}={}'.format(k, options[k]))
            opts[k] = options[k]
        else:
            raise KeyError("Option '{}' not found".format(k))

    # Bounds to np.array:
    # bounds[:, 0] - lower bounds
    # bounds[:, 1] - upper bounds
    bounds = np.array(bounds)

    # Initial population
    pop_size = opts['pop_size']
    n_params = bounds.shape[0]

    xall = np.zeros((pop_size, n_params))
    for i in range(pop_size):
        xall[i, :] = random_x(bounds)

    yall = np.full(pop_size, np.nan)

    # Add x0 if present
    if x0 is not None:
        xall[0, :] = np.array(x0)
    log.debug(f"Initial population:\n{xall}")

    # Initial evaluation
    for i in range(pop_size):
        yall[i] = fun(xall[i], *args)

    # Optimize
    tries = opts['inertia']
    count = 1        # 1, because initial pass finished
    nfev = pop_size  # pop_size, because initial pass finished
    exitmsg = ""
    ybest = yall.min()
    xbest = xall[np.argmin(yall), :]
    ybest_prev = np.array(ybest)
    xbest_prev = np.array(xbest)

    while tries > 0 and count < opts['generations']:

        # Generate new population
        xall = np.zeros((pop_size, n_params))
        xall[0, :] = xbest_prev
        for i in range(1, pop_size):
            xall[i, :] = random_x(bounds)

        # Evaluate
        for i in range(pop_size):
            yall[i] = fun(xall[i], *args)
            nfev += 1

        # Find current best
        ybest = yall.min()
        xbest = xall[np.argmin(yall), :]

        # Solution improved?
        if ybest_prev - ybest > opts['tol']:
            # Yes
            ybest_prev = ybest
            xbest_prev = xbest
            log.debug(f"Improvement = {ybest_prev - ybest:.5f}")
        else:
            # No
            tries -= 1

        # Callback
        if callback is not None:
            callback(xbest, ybest, count, *args)

        # Exit message
        if tries == 0:
            exitmsg = f"Max. number of tries ({opts['inertia']}) reached"
        if count == opts['generations']:
            exitmsg = f"Max. number of iterations ({opts['generations']}) reached"

        log.info(f"Iter.{count}: y = {ybest:.5f}")
        log.debug(f"x =\n{xbest}")

        count += 1

    # Optimization result
    res = OptRes(
        x = xbest,
        message = exitmsg,
        ng = count,
        nfev = nfev,
        fx = ybest
    )

    log.info(res)

    return res


if __name__ == "__main__":
    # Example
    logging.basicConfig(level='INFO')

    from modestga.benchmark.functions import rastrigin

    N = 50
    bounds = [(-5.12, 5.12) for i in range(N)]
    res = minimize(rastrigin, bounds, x0=None)
    print(res)



import logging
import random
import numpy as np
import matplotlib.pyplot as plt
from modestga import minimize


logging.basicConfig(filename='ga.log', level='INFO', filemode='w')

if __name__ == "__main__":
    def fun(x, *args):
        """Noisy function to be minimized"""
        return np.sum(x ** 2) + random.random()


    def callback(x, fx, ng, *args):
        """Callback function called after each generation"""

        # Print progress
        print('Generation #{}'.format(ng))
        print('    x = {}'.format(x))
        print('    fx = {}'.format(fx))

        # Save to solution history
        x_hist = args[0]
        fx_hist = args[1]
        x_hist.append(x)
        fx_hist.append(fx)


    bounds = [(0, 10) for i in range(10)]

    x_hist = list()
    fx_hist = list()
    args = (x_hist, fx_hist)

    res = minimize(fun, bounds, args=args, callback=callback, workers=1)

    # Print optimization result
    print(res)

    # Plot solution history

    fig, ax = plt.subplots(2, 1, sharex=True)
    ax[0].plot(x_hist)
    ax[0].set_title('x')
    ax[1].plot(fx_hist)
    ax[1].set_title('f(x) = np.sum(x ** 2) + random.random()')
    ax[1].set_xlabel('Generation')

    plt.show()



import logging
import random
import numpy as np


def crossover(ind1, ind2, uniform=0.5):
    # log = logging.getLogger('crossover')
    child = ind1.copy()

    rand = np.random.random(size=child.gen.size)
    child.gen = np.where(rand > uniform, child.gen, ind2.gen)

    return child


def tournament(pop, size):
    # log = logging.getLogger('tournament')

    assert size * 2 < len(pop.ind), \
        f"Tournament size ({size}) has to be lower than population ({len(pop.ind)}) // 2"

    # Form groups
    g1g2 = np.random.choice(pop.ind, size=size*2, replace=False)
    group1 = g1g2[:size]
    group2 = g1g2[size:]

    # Pick and return fittest from each group
    fit1 = np.argmin([x.val for x in group1])
    fit2 = np.argmin([x.val for x in group2])

    i1 = group1[fit1]
    i2 = group2[fit2]

    # log.debug("{}, {}".format(i1.id, i2.id))

    return i1, i2


def mutation(ind, rate, scale):
    """
    Mutate genes. `rate` controls how many genes are mutated.

    :param ind: Individual
    :param rate: float (0-1), mutation rate
    :param scale: standard deviation of the normal distribution
    :return: mutated Individual (copy)
    """
    # log = logging.getLogger('mutation')

    # Draw random value to be compared with rate
    mut = np.random.rand(ind.gen.size)

    # Draw new random genes from a normal distribution
    mut_gen = np.random.normal(loc=ind.gen, scale=scale)
    # To 0-1 range
    mut_gen = np.where(mut_gen > 1., 1., mut_gen)
    mut_gen = np.where(mut_gen < 0., 0., mut_gen)
    # Substitute genes where mut < rate
    new_gen = np.where(mut < rate, mut_gen, ind.gen)

    # log.debug(f"Old genes: {ind.gen}")
    # log.debug(f"Candidate genes {mut_gen}")
    # log.debug(f"New genes {new_gen}")

    mutind = ind.copy()
    mutind.set_genes(new_gen)

    return mutind



import logging
import time
import numpy as np
import yaml

from modestga import ga
from modestga.parallel import simple
from modestga.benchmark.functions import rastrigin

logging.basicConfig(level='INFO')


if __name__ == '__main__':
    # Test name
    tname = 'parallel_results_1'  # Change name and parameters below

    # Number of tests
    n_tests = 5

    # Number of dimensions
    n_dims = 128

    # Workers
    n_workers = [1, 2, 3, 4, 5, 6, 7, 8]

    # Functions to test
    functions = [
        ga.minimize
        # simple.minimize
    ]

    # Options
    options = {
        'generations': 1000,  # Max. number of generations
        'pop_size': 1000,      # Population size
        'tol': 1e-3           # Solution tolerance
    }

    # Results
    res = dict()

    for i in range(n_tests):
        k1 = f'run={i}'
        res[k1] = dict()

        for nw in n_workers:
            k2 = f'nworkers={nw}'
            res[k1][k2] = dict()

            for fun in functions:
                fname = fun.__module__.split('.')[-1]
                k3 = f'method={fname}'
                res[k1][k2][k3] = dict()

                opts = options.copy()

                bounds = [(-5.12, 5.12) for i in range(n_dims)]
                t0 = time.perf_counter()
                out = fun(rastrigin, bounds, x0=None, options=opts, workers=nw)
                elapsed_t = time.perf_counter() - t0

                res[k1][k2][k3]['f(x)'] = float(out.fx)
                res[k1][k2][k3]['x'] = out.x.tolist()
                res[k1][k2][k3]['nfev'] = int(out.nfev)
                res[k1][k2][k3]['ng'] = int(out.ng)
                res[k1][k2][k3]['message'] = str(out.message)
                res[k1][k2][k3]['time'] = elapsed_t

                with open(f'modestga/benchmark/results/{tname}.yaml', 'w') as yaml_file:
                    yaml.dump(res, yaml_file)

    print('Finished!')



import os
import cloudpickle
import numpy as np
import logging

from modestga import operators
from modestga import population
from modestga import individual
from modestga.ga import norm


def parallel_pop(pipe,
                 pickled_fun,
                 args,
                 bounds,
                 pop_size,
                 trm_size,
                 xover_ratio,
                 mut_rate,
                 end_event):
    """Subpopulation used in parallel GA."""
    log = logging.getLogger(name=f'parallel_pop[PID={os.getpid()}]')
    log.debug("Starting process")

    # Unpickle function
    fun = cloudpickle.loads(pickled_fun)

    # Initialize population
    pop = population.Population(pop_size, bounds, fun, evaluate=False)

    while not end_event.is_set():
        # Check if there's some data
        if pipe.poll(0.01):
            # Get data
            try:
                data = pipe.recv()
            except EOFError:
                break
            scale = data['scale']
            pop.set_genes(data['genes'])
            pop.set_fx(data['fx'])

            # Generate children
            children = list()
            fx = list()

            while len(children) < pop_size:
                #Cross-over
                i1, i2 = operators.tournament(pop, trm_size)
                child = operators.crossover(i1, i2, xover_ratio)

                # Mutation
                child = operators.mutation(child, mut_rate, scale)

                # Evaluate f(x)
                child.evaluate()

                # Add to children
                children.append(child)
                fx.append(child.val)

            # Return data (new genes) to the main process
            pop.ind = children
            data = dict()
            data['genes'] = pop.get_genes()
            data['fx'] = fx
            pipe.send(data)

    pipe.close()



import logging
import numpy as np
import matplotlib.pyplot as plt
from modestga import minimize
from modestga.benchmark.functions import rastrigin as fun


logging.basicConfig(filename='ga.log', level='INFO', filemode='w')

if __name__ == "__main__":
    # Number of dimensions
    N = 50

    def callback(x, fx, ng, *args):
        """Callback function called after each generation"""
        # Print progress
        print('Generation #{}'.format(ng))
        print('    x = {}'.format(x))
        print('    fx = {}'.format(fx))
        # Save to solution history
        x_hist = args[0]
        fx_hist = args[1]
        x_hist.append(x)
        fx_hist.append(fx)


    bounds = [(-5.12, 5.12) for i in range(N)]

    x_hist = list()
    fx_hist = list()
    args = (x_hist, fx_hist)

    options = {
        'generations': 500,
    }

    res = minimize(fun, bounds, args=args, callback=callback, options=options)

    # Print optimization result
    print(res)

    # Plot solution history
    fig, ax = plt.subplots(2, 1, sharex=True)
    ax[0].plot(x_hist, alpha=0.1, color='k')
    ax[0].set_title('x')
    ax[1].plot(fx_hist, color='k')
    ax[1].set_title('f(x) = RASTRIGIN FUNC.')
    ax[1].set_xlabel('Generation')

    plt.show()



import logging
import time
import multiprocessing
import os
import random
import numpy as np
import cloudpickle
from modestga import individual
from modestga import operators
from modestga import population


def norm(x, bounds):
    """
    Normalize `x` with respect to `bounds`.

    :param x: 1D array
    :param bounds: tuple of tuples, lower and upper bounds
    :return: 1D array, values in the range [0, 1]
    """
    n = np.array(
        [(v - b[0]) / (b[1]- b[0]) for v, b in zip(x, bounds)]
    )
    if (n < 0).any() or (n > 1).any():
        msg = "x outside bounds:\n"
        msg += "x={}\n".format(x)
        msg += "bounds={}".format(bounds)
        logging.error(msg)
        raise ValueError(msg)
    return n


def minimize(fun, bounds, x0=None, args=(), callback=None, options={}, workers=None):
    """Minimizes `fun` using Genetic Algorithm.

    If `x0` is given, the initial population will contain one individual
    based on `x0`. Otherwise, all individuals will be random.

    `fun` arguments: `x`, `*args`.

    `callback` arguments: `x`, `fx`, `ng`, `*args`.
    `fx` is the function value at the generation `ng`.

    The default options are::

        options = {
            'generations': 1000,    # Max. number of generations
            'pop_size': 100 * workers,  # Population size
            'mut_rate': 0.01,       # Mutation rate
            'trm_size': 20,         # Tournament size
            'tol': 1e-3,            # Solution tolerance
            'inertia': 100,         # Max. number of non-improving generations
            'xover_ratio': 0.5      # Crossover ratio
        }

    Returns an optimization result object with the following attributes:
    - x - numpy 1D array, optimized parameters,
    - message - str, exit message,
    - ng - int, number of generations,
    - fx - float, final function value.

    :param fun: function to be minimized
    :param bounds: tuple, parameter bounds
    :param x0: numpy 1D array, initial parameters
    :param args: tuple, positional arguments to be passed to `fun` and to `callback`
    :param callback: function, called after every generation
    :param options: dict, GA options
    :param workers: int, number of processes to use (will use all CPUs if None)
    :return: OptRes, optimization result
    """
    log = logging.getLogger(name='minimize(GA)')
    log.info('Start minimization')

    np.set_printoptions(precision=3)

    # Assign number of workers
    if workers is None:
        workers = os.cpu_count() - 1

    # Function evaluation counter
    nfev = 0

    # Options
    opts = {
        'generations': 1000,    # Max. number of generations
        'pop_size': 100 * workers,  # Population size
        'mut_rate': 0.01,       # Mutation rate
        'trm_size': 20,         # Tournament size
        'tol': 1e-3,            # Solution tolerance
        'inertia': 100,         # Max. number of non-improving generations
        'xover_ratio': 0.5      # Crossover ratio
    }

    for k in options:
        if k in opts:
            opts[k] = options[k]
        else:
            raise KeyError("Option '{}' not found".format(k))

    # Auto-adjust rules
    if ('pop_size' not in options) and (opts['pop_size'] < workers * 4):
        opts['pop_size'] = workers * 4

    if ('trm_size' not in options) and \
       (opts['trm_size'] > opts['pop_size'] // (workers * 4)):
        init_trm_size = opts['trm_size']
        opts['trm_size'] = opts['pop_size'] // (workers * 4)
        msg = (
            "Tournament size decreased due to small population: "
            f"{init_trm_size} -> {opts['trm_size']}"
        )
        log.warning(msg)

    # Assertions (detect incorrect settings)
    assert opts['pop_size'] >= (workers * 4), \
        f"Population size ({opts['pop_size']}) should be " + \
        "at least 4x larger than " + \
        f"the number of workers ({workers})"

    assert opts['trm_size'] < (opts['pop_size'] // workers), \
        f"Tournament size ({opts['trm_size']}) has to be smaller " + \
        "than population divided by number of workers " + \
        f"({opts['pop_size']}/{workers})"

    # Multiprocessing
    if workers <= 1:
        # Single process
        parallel = False
        processes = None
        pipes = None
        end_event = None
        subpop_size = None

    else:
        # Parallel processing initialization
        logging.debug(f"Using multiprocessing, workers={workers}")
        from modestga.parallel.full import parallel_pop

        parallel = True
        processes = list()
        pipes = list()
        end_event = multiprocessing.Event()
        subpop_size = opts['pop_size'] // workers

        logging.debug(f"Subpopulation size = {subpop_size}")

        for i in range(workers):
            pipe = multiprocessing.Pipe(duplex=True)
            pipe_to = pipe[0]
            pipe_from = pipe[1]
            p = multiprocessing.Process(
                target=parallel_pop,
                name=f"Subpopulation-{i}",
                args=(
                    pipe_from,
                    cloudpickle.dumps(fun),
                    args,
                    bounds,
                    subpop_size,
                    opts['trm_size'],
                    opts['xover_ratio'],
                    opts['mut_rate'],
                    end_event
                )
            )
            p.start()
            processes.append(p)
            pipes.append(pipe_to)

    # Initialize population
    pop = population.Population(opts['pop_size'], bounds, fun, args=args, evaluate=True)
    nfev += len(pop.ind)

    # Add user guess if present
    if x0 is not None:
        x0 = np.array(x0)
        pop.ind[0] = individual.Individual(
            genes=norm(x0, bounds),
            bounds=bounds,
            fun=fun,
            args=args
        )
        nfev += 1

    # Loop over generations
    ng = 0
    nstalled = 0
    vprev = None
    exitmsg = None
    scale = 0.33
    mut_rate = opts['mut_rate']

    for gi in range(opts['generations']):
        ng += 1

        # Adaptive mutation parameters
        if nstalled > (opts['inertia'] // 3):
            scale *= 0.75                                   # Search closer to current x
            mut_rate /= 1. - 1. / (len(bounds) + 1.)        # Mutate more often
            mut_rate = 0.5 if mut_rate > 0.5 else mut_rate  # But not more often than 50%

        # Fill other slots with children
        if not parallel:
            # Single process

            # Initialize children
            children = list()

            # Elitism
            children.append(pop.get_fittest())

            while len(children) < opts['pop_size']:
                #Cross-over
                i1, i2 = operators.tournament(pop, opts['trm_size'])
                child = operators.crossover(i1, i2, opts['xover_ratio'])

                # Mutation
                child = operators.mutation(child, mut_rate, scale)

                # Evaluate f(x)
                child.evaluate()
                nfev += 1

                # Add to children
                children.append(child)

            # Update population with new individuals
            pop.ind = children
        else:
            # Parallel processing
            data_to = list()
            data_from = list()

            # Divide genes among subpopulation
            all_genes = pop.get_genes()
            all_fx = pop.get_fx()
            subpop_genes = list()
            subpop_fx = list()
            for i in range(workers):
                subpop_genes.append(list())
                subpop_fx.append(list())
                for j in range(subpop_size):
                    subpop_genes[i].append(all_genes[i * subpop_size + j])
                    subpop_fx[i].append(all_fx[i * subpop_size + j])

            # Send data to workers
            for i in range(workers):
                data_to.append(dict())
                data_to[i]['scale'] = scale
                data_to[i]['genes'] = subpop_genes[i]
                data_to[i]['fx'] = subpop_fx[i]
                pipes[i].send(data_to[i])

            # Receive data from workers
            while len(data_from) < workers:
                for i in range(workers):
                    if pipes[i].poll(0.001):
                        data_from.append(pipes[i].recv())

            # Extract genes and function values
            new_genes = list()  # List (1 per subpop) of lists (1 per ind) of arrays
            new_fx = list()     # List (1 per subpop) of lists (1 per ind) of floats
            for d in data_from:
                new_genes.extend(d['genes'])
                new_fx.extend(d['fx'])
            nfev += len(new_fx)

            # Aggregate individuals of all subpopulations
            new_ind = list()
            for i in range(len(new_genes)):
                new_ind.append(
                    individual.Individual(
                        new_genes[i], bounds, fun, args=args, val=new_fx[i]
                    )
                )

            # Elitism (replace random individual)
            new_ind[random.randint(0, len(new_ind) - 1)] = pop.get_fittest()

            # Put individuals in a random order
            # (otherwise subpopulation don't exchange genes)
            random.shuffle(new_ind)

            # Update individuals
            pop.ind = new_ind

        # Tolerance check
        fittest = pop.get_fittest()
        if vprev is None:
            vprev = fittest.val
        elif abs(vprev - fittest.val < opts['tol']):
            vprev = fittest.val
            nstalled += 1
        else:
            vprev = fittest.val
            nstalled = 0

        log.info(f'ng = {gi}, nfev = {nfev}, f(x) = {fittest.val}')

        # User callback function
        if callback is not None:
            x = fittest.get_estimates()
            fx = fittest.val
            callback(x, fx, ng, *args)

        # Break if successful, i.e. f(x) = 0
        if fittest.val < opts['tol']:
            exitmsg = "Solution found, f(x) < 0 + tol"
            break

        # Break if stalled
        if nstalled >= opts['inertia']:
            exitmsg = \
                "Solution improvement below tolerance for {} generations" \
                .format(nstalled)
            break

    if ng == opts['generations']:
        exitmsg = "Maximum number of generations ({}) reached" \
            .format(opts['generations'])

    # Send message to subprocesses that the optimization is finished
    if parallel:
        end_event.set()
        for proc, pipe in zip(processes, pipes):
            log.debug(f"Closing {pipe} and joining {proc}")
            pipe.close()
            proc.join()

    # Optimization result
    fittest = pop.get_fittest()
    res = OptRes(
        x = fittest.get_estimates(),
        message = exitmsg,
        ng = ng,
        nfev = nfev,
        fx = fittest.val
    )

    return res


class OptRes:
    """
    Optimization result.

    Instance attributes:
    - x - numpy 1D array, optimized parameters
    - message - str, exit message
    - nfev - int, number of function evaluations
    - ng - int, number of generations
    - fx - float, final function value
    """
    def __init__(self, x, message, ng, nfev, fx):
        self.x = x
        self.message = message
        self.ng = ng
        self.nfev = nfev
        self.fx = fx

    def __str__(self):
        s = "Optimization result:\n"
        s += "====================\n"
        s += "x = {}\n".format(self.x)
        s += "message = {}\n".format(self.message)
        s += "ng = {}\n".format(self.ng)
        s += "nfev = {}\n".format(self.nfev)
        s += "fx = {}\n".format(self.fx)
        return s


if __name__ == "__main__":
    # Example
    logging.basicConfig(
        level='DEBUG',
        filemode='w',
        format="[%(processName)s][%(levelname)s] %(message)s"
    )
    from modestga.benchmark.functions import rastrigin
    fun = rastrigin
    bounds = [(-5.12, 5.12) for i in range(128)]
    options = {
        'generations': 10,
        'pop_size': 500,
        'tol': 1e-3
    }
    def callback(x, fx, ng, *args):
        """Callback function called after each generation"""
        print(f"\nCallback example:\nx=\n{x}\nf(x)={fx}\n")

    t0 = time.perf_counter()
    res = minimize(fun, bounds, callback=callback, options=options, workers=4)
    print(res)
    print(f"Time: {time.perf_counter() - t0}")



import numpy as np
import time


def rastrigin(x, *args):
    """Rastrigin function to be minimized.

    Global minimum y=0 at x=[0, 0, ..., 0].
    `x` can have any size.
    """
    A = 100
    n = x.size
    y = A * n + np.sum(x ** 2 - A * np.cos(2 * np.pi * x))
    return y

def rastrigin_delayed(x, *args):
    """Rastrigin function to be minimized.

    Global minimum y=0 at x=[0, 0, ..., 0].
    `x` can have any size.
    """
    time.sleep(0.01)
    A = 100
    n = x.size
    y = A * n + np.sum(x ** 2 - A * np.cos(2 * np.pi * x))
    return y



from setuptools import setup
import setuptools

setup(name='modestga',
      version='0.5.7',
      description='Genetic Algorithm minimization',
      url='https://github.com/krzysztofarendt/modestga',
      keywords='genetic algorithm optimization',
      author='Krzysztof Arendt',
      author_email='krzysztof.arendt@gmail.com',
      license='BSD',
      platforms=['Windows', 'Linux'],
      packages=setuptools.find_packages(),
      include_package_data=True,
      install_requires=[
          'numpy>=1.18.2',
          'matplotlib>=3.2.1',
          'scipy>=1.4.1',
          'cloudpickle'
      ],
      classifiers = [
          'Programming Language :: Python :: 3',
          'Topic :: Scientific/Engineering',
          'License :: OSI Approved :: BSD License'
      ]
  )



import logging
import numpy as np
from modestga.individual import Individual


class Population():

    def __init__(self, size, bounds, fun, args=(), genes=None, evaluate=True):
        self.log = logging.getLogger('Population')

        self.ind = list()

        if genes is None:
            genes = np.random.rand(len(bounds))

        if evaluate is True:
            val = None
        else:
            val = np.nan

        for i in range(size):
            self.ind.append(
                Individual(
                    genes=np.random.rand(len(bounds)),
                    bounds=bounds,
                    fun=fun,
                    args=args,
                    val=val
                )
            )

    def set_genes(self, genes):
        for i, g in enumerate(genes):
            self.ind[i].gen = g

    def set_fx(self, fx):
        for i, y in enumerate(fx):
            self.ind[i].val = y

    def get_genes(self):
        genes = [x.gen for x in self.ind]
        return genes

    def get_fx(self):
        fx = [x.val for x in self.ind]
        return fx

    def get_fittest(self):
        fittest = None
        min_fun = None

        for i in self.ind:
            if min_fun is None:
                min_fun = i.val
                fittest = i
            elif i.val < min_fun:
                min_fun = i.val
                fittest = i

        return fittest

    def __str__(self):
        s = ''
        for i in self.ind:
            s += '{}\n'.format(i)
        return s
